'''
Coordinate transforms models and concrete representations.

This package contains facilities to represent coordinate-transform objects, with
explicit semantics. The subpackages `kgprim.ct.repr` and `kgprim.ct.backend`
allow to get the concrete matrix representation of a transform.

The module `kgprim.ct.frommotions` allows to compute the coordinate transform(s)
associated with a rigid motion model from the module `kgprim.motions`.

See the file `sample/kgprim/ct/sample.py`:

```python
.. include:: ../../../sample/kgprim/ct/sample.py
```

'''



import logging
import numpy
import sympy as sym

import kgprim.values as numeric_argument
import kgprim.ct.metadata as metadata

logger = logging.getLogger(__name__)


class SymbolicMixin:
    def sin(self, arg):
        return sym.sin(arg)

    def cos(self, arg):
        return sym.cos(arg)

    def identity(self):
        return sym.eye( self.matrixSize )

    def matrix_repr(self, ct):
        mx = self.identity()
        for p in ct.primitives :
            if isinstance(p.amount, float) :
                mx = mx @ self.matrix[p.kind](p.axis, p.polarity, p.amount)
            else:
                if not isinstance(p.amount, numeric_argument.Expression) :
                    raise RuntimeError("Unknown type for the 'amount' attribute of a primitive transform of '{0}'".format(str(ct)))
                mx = mx @ self.matrix[p.kind](p.axis, p.polarity, p.amount.expr)
        return MyMx(ct, mx)




class MyMx:
    def __init__(self, ctr, mx):
        self.mx  = mx
        self.ctr = ctr
        varss, pars, consts = metadata.symbolicArgumentsOf(ctr)
        self.variables = varss
        self.parameters= pars
        self.constants = consts

        self.setParametersValue( {p : (p.defaultValue or 0) for p in self.parameters } )

    def setParametersValue(self, values):
        if len(values.keys()) != len(self.parameters) :
            logger.warn("The count of given values does not match the count of parameters")
        subs = { par.symbol : values[par] for par in values.keys() }
        self.mxNoParams = self.mx.subs( subs )
        self.eval = sym.lambdify([v.symbol for v in self.variables if v is not None], self.mxNoParams, 'numpy')

    def setVariablesValue(self, **kwargs):
        if 'valueslist' in kwargs:
            values = kwargs['valueslist']
            if len(values) != len(self.variables) :
                logger.warning("The length of the values list does not match the variables list")
            return self.eval( *values )
        elif 'valuesdict' in kwargs:
            values = kwargs['valuesdict']
            if values.keys() != self.variables.keys() :
                logger.warning("The given values do not account for all the variables of this matrix")
            replacements = { var.symbol : values[var] for var in values }
            return sym.matrix2numpy(self.mxNoParams.evalf( subs=replacements ), dtype=numpy.float64)
        else:
            logger.warning("Unrecognized parameter, skipping")

    # The following make sure that this type behaves similarly to the 'mx'
    # member, which is a Sympy matrix. Note however that this class is not meant
    # for algebra, which should be redirected explicitly to the mx member.
    def __getattr__(self, name):
        return getattr(self.mx, name)
    def __getitem__(self, key):  # this one is for subscripts [i,j]
        return self.mx[key]

#     def __matmul__(self, other): # we don't do this, as the point of this class is not to do algebra


    def __str__(self):
        return self.mx.__str__()



import kgprim.ct.repr.mxcommon as common


class RotationMatrixMixin:
    @property
    def matrixSize(self):
        return 3

    def setRotation(self, axis, polarity, mx, s, c):
        common.setRot[axis][polarity](mx, s, c)

    def setTranslation(self, axis, polarity, mx, length):
        pass

class HCoordinatesMixin:
    @property
    def matrixSize(self):
        return 4

    def setRotation(self, axis, polarity, mx, s, c):
        common.setRot[axis][polarity](mx, s, c)

    def setTranslation(self, axis, polarity, mx, length):
        common.setTr[axis][polarity](mx, length)


'''
Functions to convert rigid-motion models (`kgprim.motions`) into
coordinate-transform models.

A sequence of rigid motions specifies the relative pose between two
Cartesian frames A and B. A corresponding coordinate-transform is an object that
maps between A/B-coordinate vectors.
'''


import logging
from kgprim.ct import models as ct
from kgprim import motions as mot

logger = logging.getLogger(__name__)


ctOnRight = ct.TransformPolarity.movedFrameOnTheRight
ctOnLeft  = ct.TransformPolarity.movedFrameOnTheLeft



# The following '__xxx' methods are for internal use.

# They embed the logic to convert a sequence of motion steps to a sequence of
# primitive coordinate transforms, such that the composition of these transforms
# is the total transform associated with the given motion, **with
# `movedFrameOnTheRight` polarity**.

# Whenever the user asks for the opposite polarity (see public functions below),
# we simply reverse the motion specification, and then rely again on these
# private functions.

# We have four functions for the four combinations of the attributes that
# matter: the `mode` of the motion steps, and the `polarity` of the primitive
# transforms. Again, the `polarity` of the desired composite transforms is dealt
# with in the public conversion function (`toCoordinateTransform`), where the
# motion specification is possibly inverted before calling these functions.

# So, yes, in principle, we actually have three binary attributes required for
# unambiguous specification, and thus 8 cases in total.

def __currentRight(motion):
    return [ ct.PrimitiveCTransform(m, ctOnRight) for m in motion.steps ]

def __currentLeft(motion):
    return [ ct.PrimitiveCTransform( -m , ctOnLeft)  for m in motion.steps ]

def __fixedRight(motion):
    a = [ ct.PrimitiveCTransform(tr, ctOnRight) for tr in motion.translations()]
    b = [ ct.PrimitiveCTransform(tr, ctOnRight) for tr in reversed(motion.rotations())]
    return a + b

def __fixedLeft(motion):
    a = [ ct.PrimitiveCTransform( -tr, ctOnLeft) for tr in motion.translations()]
    b = [ ct.PrimitiveCTransform( -tr, ctOnLeft) for tr in reversed(motion.rotations())]
    return a + b

# collect the handlers above in a dictionary, for vectorization
__motionToTransformsSequence = {
    ctOnRight : {
        mot.MotionSequence.Mode.currentFrame : __currentRight,
        mot.MotionSequence.Mode.fixedFrame   : __fixedRight
    },
    ctOnLeft : {
        mot.MotionSequence.Mode.currentFrame : __currentLeft,
        mot.MotionSequence.Mode.fixedFrame   : __fixedLeft
    }
}



def __make_ct(left_frame, right_frame, motion_sequences, primitives_polarity):
    primitives = []
    for sequence in motion_sequences :
        primitives += __motionToTransformsSequence[primitives_polarity][sequence.mode](sequence)
    return ct.CoordinateTransform(left_frame, right_frame, primitives)



def toCoordinateTransform(
        poseSpec,
        right_frame = None,
        polarity = ctOnRight,
        primitives_polarity = ctOnRight):
    '''
    Convert a `kgprim.motions.PoseSpec` into a
    `kgprim.ct.models.CoordinateTransform` object.

    The pose of T (target) relative to R (reference) is converted by default to
    the transform from T coordinates to R coordinates, R_X_T. To get
    the opposite transform, pass `right_frame` equal to R; alternatively, pass
    `polarity` equal to `movedFrameOnTheLeft` (see
    `kgprim.ct.models.TransformPolarity`).

    If `right_frame` is equal to one of the frames referenced by the pose, the
    `polarity` argument is ignored.
    '''
    #TODO checks on the arguments
    onTheRight = True
    if right_frame != None :
        onTheRight = (right_frame==poseSpec.pose.target)
    else :
        # check the polarity argument only if `right_frame` was not passed
        onTheRight = (polarity==ctOnRight)

    if onTheRight :
        # Default case, we want the R_X_T coordinate transform
        sequences  = poseSpec.motion.sequences
        leftFrame  = poseSpec.pose.reference  # R
        rightFrame = poseSpec.pose.target     # T
    else :
        sequences = mot.reverse(poseSpec.motion).sequences
        leftFrame  = poseSpec.pose.target
        rightFrame = poseSpec.pose.reference

    return __make_ct(leftFrame, rightFrame, sequences, primitives_polarity)


def motionsToCoordinateTransforms(
        posesModel,
        whichTransforms=None,
        retModelName=None,
        primitivesPolarity=ctOnRight):
    '''
    Create a model with the requested coordinate transforms, from a pose
    specification model.

    Keyword arguments:

      - `posesModel`: an instance of `kgprim.motions.PosesSpec` - the source
        information
      - `whichTransforms`: a list of `kgprim.ct.models.CoordinateTransformPlaceholder`,
        which specifies which transforms should be calculated
      - `retModelName`: the name (string) of the returned object; defaults to the
        same name of the input model
      - `primitivesPolarity`: one of `kgprim.ct.models.TransformPolarity`;
        ignore if in doubt

    Return a `CTransformsModel` object, whose field `transforms` is a list
    ordered like the input list `whichTransforms`. If `whichTransforms` is
    None, the default transforms will be returned: these are the A_X_B
    transforms for each input motion A-->B.

    '''
    retModelName = retModelName or posesModel.name
    if whichTransforms is None :
        return __convertModel(posesModel, retModelName)
    transforms = []
    inspector = mot.ConnectedFramesInspector(posesModel)
    for whichone in whichTransforms :
        poseSpec = inspector.getPoseSpec(targetFrame=whichone.leftFrame, referenceFrame=whichone.rightFrame)
        if poseSpec is None :
            logger.warning("Could not determine relative pose between '{0}' and '{1}'".format(whichone.leftFrame, whichone.rightFrame))
        else :
            myct = toCoordinateTransform(poseSpec, right_frame=whichone.rightFrame, primitives_polarity=primitivesPolarity)
            transforms.append(myct)
    return ct.CTransformsModel(retModelName, transforms)


def __convertModel(
        posesModel, retModelName,
        mode=ctOnRight,
        primitivesPolarity=ctOnRight):
    '''
    Create the default coordinate transforms model from a pose specification
    model.

    Returns a `kgprim.ct.models.CTransformsModel`.

    For each motion A-->B in `posesModel`, the returned transforms model
    will contain either the A_X_B or B_X_A transform, depending on `mode`.
    '''

    transforms = []
    for pose in posesModel.poses :
        transforms.append(
            toCoordinateTransform(pose, None, mode, primitives_polarity=primitivesPolarity) )

    return ct.CTransformsModel(retModelName, transforms)






'''
Matrix representations for coordinate transforms.

This module contains a few functors that create a matrix representation of the
given `kgprim.ct.models.CoordinateTransform` model. The available
representations are:
  - pure rotation matrix (possible translation components in the transform are
    discarded)
  - 4x4 matrix for homogeneous coordinates vectors
  - 6x6 matrix for spatial motion vectors
  - 6x6 matrix for spatial force vectors

For any matrix, one can choose between two concrete backends for the matrix
data: numeric and symbolic (using respectively Numpy and Sympy).
The symbolic option is required whenever the coordinate transform depends on at
least one non-constant argument, like a `kgprim.values.Variable` or a
`kgprim.values.Parameter`.

For example:

```python
import kgprim.ct.repr.mxrepr as mxrepr

H = mxrepr.hCoordinatesNumeric( ct )
R = mxrepr.rotationMatrixNumeric( ct )
M = mxrepr.spatialMotionSymbolic( ct )
```

Please see `test/ct/sample.py` in the project repository for a more complete
example.
'''

from kgprim.ct.repr import mxcommon
from kgprim.ct.repr import homogeneous
from kgprim.ct.repr import spatial
from kgprim.ct.backend.numeric  import NumericMixin
from kgprim.ct.backend.symbolic import SymbolicMixin

from enum import Enum

class MatrixRepresentation(Enum):
    '''
    Enumeration of the matrix representations available from this module
    '''

    homogeneous = 0
    spatial_motion = 1
    spatial_force = 2
    pure_rotation = 3


class MatrixRepresentationMixin:
    def __init__(self, **kwds):
        super().__init__(**kwds)
        self.matrix = {
            mxcommon.ROT : self.rotation,
            mxcommon.TR  : self.translation
        }

    def rotation(self, axis, polarity, angle):
        mx = self.identity()
        s  = self.sin(angle);
        c  = self.cos(angle);
        self.setRotation(axis, polarity, mx, s, c)
        return mx

    def translation(self, axis, polarity, length):
        mx = self.identity()
        self.setTranslation(axis, polarity, mx, length)
        return mx

    # make the object look like a functor, returning the matrix representation
    def __call__(self, ct):
        return self.matrix_repr(ct) # this is defined in the backend mixins


# Compose the mixins to get concrete types that can produce a matrix
# representation of a coordinate transform:

class RotationMatrixNumeric (MatrixRepresentationMixin, NumericMixin , homogeneous.RotationMatrixMixin): pass
class RotationMatrixSymbolic(MatrixRepresentationMixin, SymbolicMixin, homogeneous.RotationMatrixMixin): pass

class HCoordinatesNumeric (MatrixRepresentationMixin, NumericMixin , homogeneous.HCoordinatesMixin): pass
class HCoordinatesSymbolic(MatrixRepresentationMixin, SymbolicMixin, homogeneous.HCoordinatesMixin): pass

class SpatialMotionNumeric (MatrixRepresentationMixin, NumericMixin , spatial.MotionVectorMixin): pass
class SpatialMotionSymbolic(MatrixRepresentationMixin, SymbolicMixin, spatial.MotionVectorMixin): pass

class SpatialForceNumeric (MatrixRepresentationMixin, NumericMixin , spatial.ForceVectorMixin): pass
class SpatialForceSymbolic(MatrixRepresentationMixin, SymbolicMixin, spatial.ForceVectorMixin): pass

rotationMatrixNumeric  = RotationMatrixNumeric ()
rotationMatrixSymbolic = RotationMatrixSymbolic()

hCoordinatesNumeric    = HCoordinatesNumeric   ()
hCoordinatesSymbolic   = HCoordinatesSymbolic  ()

# The spatial motion representation defaults to the convention with rotational
# coordinates on top of the matrix. To use the other convention, pass the
# keyword argument, as in:
#
# obj = SpatialMotionNumeric(spatialCoordinatesConvention = spatial.CoordinatesConvention.translationOnTop)

spatialMotionNumeric   = SpatialMotionNumeric  ()
spatialMotionSymbolic  = SpatialMotionSymbolic ()

spatialForceNumeric    = SpatialForceNumeric   ()
spatialForceSymbolic   = SpatialForceSymbolic  ()

symbolic = {
    MatrixRepresentation.homogeneous    : hCoordinatesSymbolic,
    MatrixRepresentation.spatial_motion : spatialMotionSymbolic,
    MatrixRepresentation.spatial_force  : spatialForceSymbolic,
    MatrixRepresentation.pure_rotation  : rotationMatrixSymbolic
}



def constantCoefficients(symbMatrix):
    constants = []
    variables = []
    for r in range(0, symbMatrix.rows) :
        for c in range(0, symbMatrix.cols) :
            if symbMatrix[r,c].is_constant() :
                constants.append( (r,c) )
                # if symbMatrix[r,c].is_Float :
                #    floats.append( (r,c) )
            else :
                variables.append( (r,c) )
    return tuple(constants), tuple(variables)

class MatrixReprMetadata:
    '''
    Metadata of a matrix representation of a coordinate transform
    '''

    def __init__(self, coordinateTransformMetadata, matrixRepresentation,
                 reprKind):

        ccoeff, vcoeff = constantCoefficients(matrixRepresentation)
        self.variableCoefficients = vcoeff
        self.constantCoefficients = ccoeff
        self.mx = matrixRepresentation
        self.representationKind = reprKind
        self.ctMetadata = coordinateTransformMetadata

    def rows(self):   return self.mx.rows
    def cols(self):   return self.mx.cols

    def __eq__(self, rhs):
        return (isinstance(rhs, MatrixReprMetadata)
                and (self.ctMetadata.ct == rhs.ctMetadata.ct)
                and (self.mx == rhs.mx))

    def __hash__(self) :
        return 31*hash(self.ctMetadata.ct) + 93*hash(self.mx)




'''
Core primitives like `RigidBody` or Cartesian `Frame`.

All the classes are nothing more than named entities, and are meant to be used
as symbolic models. For example, a `Pose` is represented simply by two instances
of `Frame`, with two different names.
'''

class RigidBody():
    def __init__(self, n):
        '''
        Constructor arguments:
          - `n`: the desired name for this instance
        '''
        self._name = n

    @property
    def name(self):
        '''The name of this rigid body'''
        return self._name

    def __eq__(self, rhs): return isinstance(rhs, RigidBody) and self.name==rhs.name
    def __hash__(self)   : return 47 * hash(self._name)
    def __str__(self)    : return self.name


class Attachable():
    def __init__(self):
        None


class Point(Attachable):
    def __init__(self, n):
        super().__init__()
        self._name = n

    @property
    def name(self): return self._name

    def __eq__(self, rhs): return isinstance(rhs, Point) and self.name==rhs.name
    def __hash__(self)   : return 31 * hash(self._name)
    def __str__(self)    : return self.name



class Frame(Attachable):
    '''
    A Cartesian coordinate frame.
    '''

    def __init__(self, n):
        '''
        Constructor arguments:
          - `n`: the desired name for this instance
        '''
        self._name = n

    @property
    def name(self): return self._name

    def __eq__(self, rhs):
        return isinstance(rhs, Frame) and self._name == rhs._name

    def __hash__(self):
        return 97 * hash(self._name)

    def __str__(self):
        return self.name


class Pose:
    '''
    The relative pose of `target` with respect to `reference`.

    Both `target` and `reference` should be two instances of `Frame`.
    '''

    def __init__(self, target, reference):
        self.target    = target
        self.reference = reference

    def __eq__(self, rhs):
        return isinstance(rhs, Pose) and self.target==rhs.target and self.reference==rhs.reference

    def __hash__(self):
        return 43*hash(self.target) + 11*hash(self.reference)

    def __str__(self):
        return self.target.name + " wrt " + self.reference.name
    def __repr__(self):
        return self.__str__()

class Velocity():
    def __init__(self, target, reference):
        self.target    = target
        self.reference = reference

    def __eq__(self, rhs):
        return isinstance(rhs, Velocity) and self.target==rhs.target and self.reference==rhs.reference

    def __hash__(self):
        return 77*hash(self.target) + 19*hash(self.reference)

    def __str__(self):
        return "V of " + self.target.name + " wrt " + self.reference.name
    def __repr__(self):
        return self.__str__()


def poseCompose(pose1, pose2):
    if pose1.reference != pose2.target :
        raise RuntimeError("Cannot compose {0} with {1}, reference/target mismatch".format(pose1,pose2))
    return Pose(pose1.target, pose2.reference)

def velCompose(v1, v2):
    if v1.reference != v2.target:
        raise RuntimeError("Cannot compose {0} with {1}, reference/target mismatch".format(v1,v2))
    return Velocity(v1.target, v2.reference)


class Attachment():
    '''
    Models the attachment relation between an attachable entity (like a
    Frame) and a RigidBody
    '''
    def __init__(self, entity, body):
        if not isinstance(body, RigidBody) :
            raise ValueError("Things can only be attached to rigid bodies")
        if not isinstance(entity, Attachable) :
            raise ValueError("Only 'attachable' objects like Points can be attached to rigid bodies")

        self._entity = entity
        self._body   = body
        self.attrs   = {}

    @property
    def entity(self): return self._entity
    @property
    def body(self): return self._body

    # Pretend to be the entity
    def __getattr__(self, name):
        return getattr(self.entity, name)

    def __str__(self):
        return self.entity.name + " attached to " + self.body.name
    def __repr__(self):
        return self.entity.name + "@" + self.body.name


'''
Numeric (based on Numpy) or symbolic (Sympy) backends for the concrete matrix
representations of coordinate transforms.

The symbolic backend is required whenever the coordinate transform depends on
at least one non-constant argument, like a `kgprim.values.Variable` or a
`kgprim.values.Parameter`.
'''



from kgprim.motions import Axis
from kgprim.motions import MotionStep
from kgprim.ct      import models

def __rotx__a_x_b(mx, s, c):
    mx[1,1] = c;
    mx[1,2] = -s;
    mx[2,1] = s;
    mx[2,2] = c;

def __rotx__b_x_a(mx, s, c):
    mx[1,1] = c;
    mx[2,1] = -s;
    mx[1,2] = s;
    mx[2,2] = c;

def __roty__a_x_b(mx, s, c):
    mx[0,0] = c;
    mx[0,2] = s;
    mx[2,0] = -s;
    mx[2,2] = c;

def __roty__b_x_a(mx, s, c):
    mx[0,0] = c;
    mx[2,0] = s;
    mx[0,2] = -s;
    mx[2,2] = c;

def __rotz__a_x_b(mx, s, c):
    mx[0,0] = c;
    mx[0,1] = -s;
    mx[1,0] = s;
    mx[1,1] = c;

def __rotz__b_x_a(mx, s, c):
    mx[0,0] = c;
    mx[1,0] = -s;
    mx[0,1] = s;
    mx[1,1] = c;


def __tr_x__a_x_b(mx, length):
    mx[0,3] = length

def __tr_x__b_x_a(mx, length):
    mx[0,3] = -length

def __tr_y__a_x_b(mx, length):
    mx[1,3] = length

def __tr_y__b_x_a(mx, length):
    mx[1,3] = -length

def __tr_z__a_x_b(mx, length):
    mx[2,3] = length

def __tr_z__b_x_a(mx, length):
    mx[2,3] = -length


onRight = models.TransformPolarity.movedFrameOnTheRight
onLeft  = models.TransformPolarity.movedFrameOnTheLeft

X = Axis.X
Y = Axis.Y
Z = Axis.Z

setRot = {
    X : {
        onRight : __rotx__a_x_b,
        onLeft  : __rotx__b_x_a
    },
    Y : {
        onRight : __roty__a_x_b,
        onLeft  : __roty__b_x_a
    },
    Z : {
        onRight : __rotz__a_x_b,
        onLeft  : __rotz__b_x_a
    },
}

setTr = {
    X : {
        onRight : __tr_x__a_x_b,
        onLeft  : __tr_x__b_x_a
    },
    Y : {
        onRight : __tr_y__a_x_b,
        onLeft  : __tr_y__b_x_a
    },
    Z : {
        onRight : __tr_z__a_x_b,
        onLeft  : __tr_z__b_x_a
    },
}

ROT = MotionStep.Kind.Rotation
TR  = MotionStep.Kind.Translation




'''
This module contains types to represent symbolically rigid motions.
Motions are pure translations/rotations ("primitive" motions), and sequences
of them.

A sequence of rigid motions is really a specification of a relative pose
between Cartesian frames: the motions required to move from A to B identify
the pose of B relative to A. Conversely, the pose of B relative to A could be
_defined_ with a sequence of rigid motions that bring A to coincide with B.
Therefore, these docs may use the term 'motion' and 'pose-specification'
interchangeably.

Rigid motion models (relative poses) can be turned into concrete
representations such as coordinate transforms. This is in fact a typical use
case. See the package `kgprim.ct` and the module `kgprim.ct.frommotions`.

The package `motiondsl` offers a simple textual format to conveniently load
motion objects.
'''

from enum import Enum
from kgprim.core import Frame
from kgprim.core import Pose


class Axis(Enum):
    X=0
    Y=1
    Z=2

class MotionStep:
    '''
    A representation of a pure rotation/translation motion, about/along one
    Cartesian axis.

    Any instance also stores an `amount` member to quantify the magnitude of
    the step.
    '''
    class Kind(Enum):
        '''Either `Rotation` or `Translation`'''
        Rotation=0
        Translation=1

    def __init__(self, kind, axis, amount):
        self.kind   = kind
        self.axis   = axis
        self.amount = amount

    def __neg__(self):
        return MotionStep(self.kind, self.axis, - self.amount)

    __strbits = {
        Kind.Rotation    : "Rot",
        Kind.Translation : "Tr",
    }
    def __str__(self) :
        return MotionStep.__strbits[self.kind] + self.axis.name + "(" + str(self.amount) + ")"
    def __repr__(self): return self.__str__()

    def __eq__(self, rhs):
        return (isinstance(rhs, MotionStep)
               and self.kind == rhs.kind
               and self.axis == rhs.axis
               and self.amount == rhs.amount)
    def __hash__(self):
        return 7*hash(self.kind) + 13*hash(self.axis) + 17*hash(self.amount)

def translation(axis, amount): return MotionStep(MotionStep.Kind.Translation, axis, amount)
def rotation   (axis, amount): return MotionStep(MotionStep.Kind.Rotation   , axis, amount)



class MotionSequence:
    '''
    An ordered sequence of `MotionStep`s taking place according to
    the _same_ convention (see `MotionSequence.Mode`).

    A `MotionSequence` is a sequence of pure rotations/translations, in any
    order, plus an attribute telling how to interpret the sequence itself: see
    `Mode`.
    '''

    class Mode(Enum):
        '''
        The two possible ways of interpreting a sequence of motion steps.

        This is an enumeration with two items:

        - `currentFrame`: every motion step is taken with respect to the
           current, moving Cartesian frame
        - `fixedFrame`: the reference frame stands still, thus the `axis` of any
           motion step is always one of the fixed axes.
        '''
        currentFrame = 0
        fixedFrame   = 1

    def __init__(self, steps, mode=Mode.currentFrame):
        self.steps= steps
        self.mode = mode

    @property
    def sequences(self):
        '''
        A list containing only self.

        This property is implemented to emulate the behaviour of `MotionPath`, so
        that the two types can be used interchangeably.
        '''
        return [ self ]

    def translations(self):
        '''
        A list containing only the translations of this instance, in the same
        relative order.
        '''
        return [t for t in self.steps if t.kind == MotionStep.Kind.Translation]

    def rotations(self):
        '''
        A list containing only the rotations of this instance, in the same
        relative order.
        '''
        return [t for t in self.steps if t.kind == MotionStep.Kind.Rotation]



class MotionPath:
    '''
    A representation of a rigid motion, as a sequence of `MotionSequence`s.

    The point of this class is really to model a single sequence of motion
    steps (a "path") but to allow subsequences to follow a different convention
    (see `MotionSequence.Mode`).

    An instance can be constructed with a list of `MotionSequence` objects,
    or a list of `MotionPath` objects. In both cases, the objects are composed
    to form an individual descriptor.
    '''

    def __init__(self, sequences):

        # To make sure we internally store an homogeneous list of MotionSequence
        # objects, we must unwrap possible MotionPath objects given to this
        # constructor. When concatenating MotionPaths, we really want to
        # concatenate the MotionSequences stored internally. Composing
        # MotionPaths leads to another MotionPath whose sequence is the
        # composition of the sequences of the original MotionPaths

        self._sequences = []
        for motion in sequences :
            if isinstance(motion, MotionSequence) :
                self._sequences.append( motion )
            elif isinstance(motion, MotionPath ) :
                for subs in motion.sequences :
                    self._sequences.append( subs )
            else:
                raise RuntimeError("Unknonw motion type: " + str(type(motion)))

    @property
    def sequences(self):
        '''The list of `MotionSequences` this motion descriptor is comprised of.'''
        return self._sequences



def reverse(aMotion) :
    inverseSequences = []
    for seq in aMotion.sequences[::-1] :#reverse view
        if seq.mode == MotionSequence.Mode.currentFrame :
            steps = [ -step for step in seq.steps[::-1] ]
            inverseSequences.append( MotionSequence(steps, seq.mode) )
        else : # fixed-frame mode
            rot = seq.rotations()
            tr  = seq.translations()

            rot2 = [ -step for step in rot[::-1] ]
            tr2  = [ -step for step in tr[::-1] ]
            inverseSequences.append( MotionSequence(rot2, MotionSequence.Mode.fixedFrame) )
            inverseSequences.append( MotionSequence(tr2, MotionSequence.Mode.currentFrame) )

    return MotionPath(inverseSequences)


class PoseSpec:
    '''
    A relative Pose augmented with the corresponding motion description.

    A sequence of rigid motions is really a specification of a relative pose
    between Cartesian frames: the motions required to move from A to B identify
    the pose of B relative to A.
    '''

    def __init__(self, pose, motion, name=None):
        '''
        Arguments:
        - pose: the `kgprim.core.Pose` instance you want to describe
        - motion: an instance of `MotionSequence` or `MotionPath`, which lists
          the motion-steps defining the pose
        '''
        self._pose   = pose
        self._motion = motion
        self._name = name

    @property
    def pose(self): return self._pose

    @property
    def motion(self): return self._motion

    @property
    def name(self): return self._name

def inversePoseSpec(poseSpec):
    pose = Pose(target=poseSpec.pose.reference, reference=poseSpec.pose.target)
    motion = reverse(poseSpec.motion)
    name = None
    if poseSpec.name is not None:
        name = f"inverse-of-{poseSpec.name}"
    return PoseSpec(pose, motion, name)



class PosesSpec:
    '''
    A simple named container for a list of `PoseSpec` instances.
    '''

    def __init__(self, name, poses):
        self._name = name
        self._poses = poses

    def mergeModel(self, otherModel, name=None):
        if name is None:
            name = self.name + '_' + otherModel.name
        poses = self.poses.copy() # A shallow copy
        poses.extend( otherModel.poses ) # only works for lists !
        return PosesSpec(name, poses)

    @property
    def name(self):
        '''The name given to this container'''
        return self._name

    @property
    def poses(self):
        '''The list of `PoseSpec` in this container'''
        return self._poses

    #TODO operator[] and such, to make it act as the list


import networkx as nx

class ConnectedFramesInspector:
    '''
    An inspector to determine which additional poses can be induced from an
    existing PosesSpec model.

    For example, if the model contains the pose of A relative to B, and A
    relative to C, it is obviously possible to infer the relation between B and
    C.
    '''
    def __init__(self, posesModel):
        self.posesModel = posesModel
        graph = nx.DiGraph()

        for poseSpec in posesModel.poses :
            pose   = poseSpec.pose
            motion = poseSpec.motion
            tgt  = pose.target
            ref  = pose.reference
            graph.add_edge( ref, tgt, motion=motion )
            graph.add_edge( tgt, ref, motion=reverse(motion) )

        self.graph = graph


    def hasRelativePose(self, frame1, frame2):
        if (frame1 not in self.graph) or (frame2 not in self.graph):
            return False
        return nx.has_path(self.graph, frame1, frame2)

    def getPoseSpec(self, targetFrame, referenceFrame):
        if not self.hasRelativePose(targetFrame, referenceFrame) :
            return None
        path = nx.shortest_path(self.graph, target=targetFrame, source=referenceFrame)
        if len(path) < 1:
            return None
        motions = []
        for v1,v2 in zip(path, path[1:]) :
            edgedata = self.graph.edges[v1,v2]
            motions.append(edgedata['motion'])

        pose = Pose(target=targetFrame, reference=referenceFrame)
        return PoseSpec(pose=pose, motion=MotionPath(motions))
        # TODO: return the original PoseSpec, if there is one that matches the
        # given frames, instead of recreating it




'''
A DSL (Domain Specific Language) to specify rigid motions - the "MotionsDSL"

This package defines a simple language to represent objects of the
`kgprim.motions` module. That is, it defines *a* file format to specify rigid
motions and sequences of rigid motions.

A generic entry of a document of the MotionDSL has this format:

`<frame name 1> -> <frame name 2> : <spec>`

which reads: "to go from `frame1` to `frame2`, `<spec>` steps must be taken.
Here is a short sample of a complete document:

    Model MotionDSLSample

    Convention = currentFrame
    // -> all the motion-steps referenced by this model are to be interpreted
    // in the current, moving frame

    fA -> fB : rotx(0.2) try(3.1)
    // A rotation about X followed by a translation along Y.
    // Use float literals for constant amounts

    // Demonstrate the use of variables
    fE -> fF : rotx(q0)
    fF -> fG : roty(q1)

Identifiers appearing as the argument of a motion step are interpreted as
variables. The corresponding `kgprim.motions.MotionStep` object created when
loading the document will have a `kgprim.values.Variable` instance as the value
of its `amount` member.

See the file `sample/motiondsl/model.motdsl` for a slightly larger sample. See
also `sample/motiondsl/sample.py` for an example of how to read the model with
Python code.

While this package depends on the `kgprim.motions` module, the
converse is not true.

We use [textX](https://github.com/textX/textX) to create the parser of the
language.
'''



'''
A set of types to represent explicitly variables, parameters and constants.

The types do not really implement any particular behaviour, but
are mere placeholders provided for modeling purposes. Instances can be used in
alternative to floats, for numerical attributes (such as the `amount` of a
`kgprim.motions.MotionStep` object).

Conceptually, the relative rate of change of a numerical attribute determines
whether it is a variable, a parameter, or a constant. It is a modelling choice
to use a class or another.

Internally, a Sympy symbol is used, which is accessible via the `symbol`
property. The `expr` property also returns the Sympy symbol; it is provided
for uniformity with the `Expression` class, where `expr` resolves to the actual
Sympy expression.
'''

import sympy as sp

class Variable:
    '''
    An explicit representation of a named variable.

    This class is implemented as a simple wrapper of a Sympy symbol.
    '''

    def __init__(self, name, symbol=None):
        self.name_ = name
        if symbol is not None :
            self.symbol_ = symbol
        else :
            self.symbol_ = sp.Symbol(name=name)
    @property
    def name(self):
        return self.name_
    @property
    def symbol(self):
        '''The Sympy symbol used internally'''
        return self.symbol_

    @property
    def expr(self):
        return self.symbol
    @property
    def constant(self):
        return False

    def __eq__(self, rhs):
        return (isinstance(rhs, Variable)
                and (self.name == rhs.name)
                and (self.symbol  == rhs.symbol))
    def __hash__(self) :
        return 47*hash(self.name) + 93*hash(self.symbol)
    def __str__(self):
        return "var:{n}".format(n=self.name)
    def __repr__(self):
        return self.__str__()


class Parameter:
    '''
    An exlicit representation of a named parameter, with an optional default
    value.

    Like `Variable`, this class is also a simple wrapper of a Sympy symbol.

    Only the name and the symbol of the parameter influence equality check and
    hashing, NOT the default value.
    '''

    def __init__(self, name, symbol=None, defValue=None):
        self.name_   = name
        self.symbol_ = symbol or sp.Symbol(name=name)
        self.dvalue_ = defValue

    @property
    def name(self):
        return self.name_

    @property
    def symbol(self):
        '''The Sympy symbol used internally'''
        return self.symbol_
    @property
    def expr(self):
        return self.symbol
    @property
    def constant(self):
        return False

    @property
    def defaultValue(self):
        return self.dvalue_

    def __eq__(self, rhs):
        return (isinstance(rhs, Parameter)
                and (self.name == rhs.name)
                and (self.symbol  == rhs.symbol))
    def __hash__(self) :
        return 101*hash(self.name) + 11*hash(self.symbol)
    def __str__(self):
        return "param:{n}".format(n=self.name)
    def __repr__(self):
        return self.__str__()



class _ConstantSymbol(sp.NumberSymbol):
    # For internal use.
    # Inherit from sympy.NumberSymbol to create a generic symbol with a constant
    # value. Don't really know whether this is correct/complete, but it seems it
    # does the right thing. Not very much documentation out there about how to
    # create a custom constant symbol

    __slots__ = ['name', 'value']

    def __new__(cls, name, value):
        self = super().__new__(cls)
        self.name  = name
        self.value = sp.Float(value)
        return self

    def _as_mpf_val(self, prec):
        return self.value._as_mpf_val(prec)

    def _sympystr(self, printer):
        return printer.doprint(sp.Symbol(self.name))

    def __eq__(self, rhs):
        return (isinstance(rhs, _ConstantSymbol)
                and (self.name  == rhs.name)
                and (self.value == rhs.value))
    def __hash__(self) :
        return 113*hash(self.name) + 29*hash(self.value)


class Constant:
    '''
    A wrapper of a named symbol with a constant floating point value.
    '''

    def __init__(self, name, value):
        self._symbol= _ConstantSymbol(name, value)
        self._name  = name
        self._value = value

    @property
    def name(self):
        return self._name
    @property
    def symbol(self):
        '''The Sympy symbol used internally'''
        return self._symbol
    @property
    def expr(self):
        return self.symbol
    @property
    def value(self):
        return self._value
    @property
    def constant(self):
        return True

    def __eq__(self, rhs):
        return (isinstance(rhs, Constant)
                and (self._name  == rhs._name)
                and (self._value == rhs._value))
    def __hash__(self) :
        return 107*hash(self._name) + 13*hash(self._value)

class MyPI:
    '''
    A singleton representing the constant pi
    '''

    __inst = None

    def __init__(self):
        self.name_ = "pi"

    @property
    def name(self):
        return self.name_
    @property
    def symbol(self):
        '''The Sympy object used internally'''
        return sp.pi
    @property
    def expr(self):
        return self.symbol
    @property
    def constant(self):
        return True
    @property
    def value(self):
        return sp.pi.evalf()

    @staticmethod
    def instance():
        if MyPI.__inst is None:
            MyPI.__inst = MyPI()
        return MyPI.__inst

class Expression:
    '''
    A wrapper of simple expressions involving either a `Variable`, `Parameter`
    or `Constant`. The expression can have only one argument.

    The actual symbolic expression like '3 * var' or 'pi/2' is stored internally
    as a Sympy expression.
    '''

    def __init__(self, argument, sympyExpr=None):
        '''
        Parameters:
          - `argument`: an instance of either `Variable`, `Parameter` or
            `Constant`, which is the argument of the expression.
          - `sympyExpr` : the actual, full, Sympy expression; if None, this
             instance will represent the identity expression.
        The only free symbol in the given `sympyExpr` must be the same as
        `argument.symbol`, otherwise a `RuntimeError` is raised.
        '''
        if sympyExpr is not None :
            if len(sympyExpr.free_symbols) > 0: # ==0 only for pi or other constants
                if sympyExpr.free_symbols.pop() != argument.symbol :
                    raise RuntimeError('Inconsistent arguments')
            self.expression = sympyExpr
        else :
            self.expression = argument.expr
        self.argument = argument

    @property
    def expr(self):
        '''The underlying Sympy expression'''
        return self.expression
    @property
    def arg(self):
        '''The single argument of this expression.'''
        return self.argument

    def evalf(self):
        '''The floating point value of this expression, if it is constant'''
        if not self.constant() :
            raise RuntimeError('Cannot evaluate to float a non constant expression')
        return float( self.expression.evalf( subs={self.argument.symbol : self.argument.value}) )

    def constant(self):
        '''Whether this expression has a constant value or not'''
        return self.argument.constant

    def __neg__(self):
        return Expression(self.argument, -self.expr)

    def __mul__(self, rhs):
        return Expression(self.argument, self.expr*rhs)

    def __truediv__(self, rhs):
        return Expression(self.argument, self.expr/rhs)

    def __eq__(self, rhs):
        return (isinstance(rhs, Expression)
                and (self.arg == rhs.arg)
                and (self.expr== rhs.expr))
    def __hash__(self) :
        return 31*hash(self.arg) + 93*hash(self.expr)
    def __str__(self):
        return self.expr.__str__()
    def __repr__(self):
        return "expr:"+self.__str__()

    __rmul__ = __mul__



__evalf = {
    Expression: lambda E: E.evalf(),
    float:      lambda f: f
}

def toFloat(expr):
    return __evalf[expr.__class__](expr)



'''
The actual implementation of the MotionDSL (`motiondsl`), using
[textX](https://github.com/textX/textX).

The module-level member `dsl` is the default instance of the `MotionDSL` class.
Normally, you will only need this preallocated instance to load a document in
the MotionDSL format.

For example:

```python
import os, sys
import motiondsl.motiondsl as motdsl

def main():
    if len(sys.argv) > 1 :
        ifile = sys.argv[1]
    else :
        ifile = os.path.join( os.path.dirname(__file__), 'sample.motdsl')

    # Load the motions-model from the input file
    motionsModel = motdsl.dsl.modelFromFile(ifile)

    print("Motions (poses) model name: ", motionsModel.name)

    # Convert the motions-model to a pose-specification model
    posesModel = motdsl.toPosesSpecification(motionsModel)
```

See also the files in the `sample/motiondsl` folder.
'''

import sympy as sp
import os
import textx
import kgprim.core as primitives
from kgprim.values import Variable
from kgprim.values import Parameter
from kgprim.values import Constant
from kgprim.values import MyPI
from kgprim.values import Expression
from kgprim.motions import Axis
from kgprim.motions import MotionStep
from kgprim.motions import MotionSequence
from kgprim.motions import PoseSpec
from kgprim.motions import PosesSpec


def _argValue(arg):
    if type(arg) == float :
        return arg
    else:
        return Expression(arg)

class MotionDSL:
    def __init__(self):
        here = os.path.dirname(os.path.abspath(__file__))
        self.mm = textx.metamodel_from_file(here+"/motiondsl.tx", auto_init_attributes=False)
        # disable auto-init so that optional attributes (like the default value
        # of a parameter) will be 'None' when not specified in the input model

        obj_processors = {
            'PILiteral': lambda _ : MyPI.instance(),
            'Variable' : lambda x : Variable(name=x.name),
            'Parameter': lambda x : self._instantiateParameter(x),
            'UserConstant': lambda x : Constant(name=x.name, value=x.value),
            'RefToConstant': lambda x : Constant(name=x.actual.name, value=x.actual.value),

            # the following result in a float or an Expression object
            'PlainExpr' : lambda e: (-1 if e.minus else 1) * _argValue( e.arg ),
            'MultExpr'  : lambda e: e.mult * _argValue( e.arg ),
            'DivExpr'   : lambda e: (-1 if e.minus else 1)* _argValue( e.arg ) / e.div,

            'Trx' : lambda m : MotionStep(MotionStep.Kind.Translation, Axis.X, m.expr),
            'Try' : lambda m : MotionStep(MotionStep.Kind.Translation, Axis.Y, m.expr),
            'Trz' : lambda m : MotionStep(MotionStep.Kind.Translation, Axis.Z, m.expr),
            'Rotx': lambda m : MotionStep(MotionStep.Kind.Rotation,    Axis.X, m.expr),
            'Roty': lambda m : MotionStep(MotionStep.Kind.Rotation,    Axis.Y, m.expr),
            'Rotz': lambda m : MotionStep(MotionStep.Kind.Rotation,    Axis.Z, m.expr)
        }
        self.mm.register_obj_processors( obj_processors )
        self._resetState()

    def _resetState(self):
        self.paramInstances = {}
        self.paramDefValues = {}

    def _instantiateParameter(self, paramFromModel):
        # all the parameters with the same name (i.e. references to the same
        #  model parameter) get the default value which was encountered first

        name     = paramFromModel.name
        plist    = self.paramInstances.setdefault(name, [])
        defvalue = self.paramDefValues.get(name)
        if defvalue is None and paramFromModel.defvalue is not None:
            defvalue = paramFromModel.defvalue
            self.paramDefValues[name] = defvalue
            for p in plist:
                p.dvalue_ = defvalue

        newp = Parameter(name=name, defValue=defvalue)
        plist.append(newp)
        return newp


    def modelFromFile(self, file):
        '''
        The MotionDSL model represented by the given conforming document.

        Arguments:
          - `file`: path of a MotionDSL document (i.e. a text file)
        '''
        self._resetState()
        return self.mm.model_from_file(file)


    def modelFromText(self, text):
        self._resetState()
        return self.mm.model_from_str(text)


dsl = MotionDSL()

__mode_map = {
    "currentFrame" : MotionSequence.Mode.currentFrame,
    "fixedFrame"   : MotionSequence.Mode.fixedFrame
}


def toPosesSpecification(dslmodel):
    '''
    Convert a MotionDSL model into a `kgprim.motions.PosesSpec` instance.

    One typically gets a MotionDSL model by loading it from an input document;
    see `MotionDSL.modelFromFile()`
    '''
    mode = __mode_map[ dslmodel.convention ]
    poses = []
    for m in dslmodel.motions :
        ref  = primitives.Frame(m.start.name)
        tgt  = primitives.Frame(m.end.name)
        pose = primitives.Pose(target=tgt, reference=ref)
        motSeq = MotionSequence( m.primitiveMotions, mode)
        poses.append( PoseSpec(pose=pose, motion=motSeq, name=m.userName) )

    return PosesSpec(name=dslmodel.name, poses=poses)



__ser_map = {
    MotionStep.Kind.Translation : {
        Axis.X : 'trx',
        Axis.Y : 'try',
        Axis.Z : 'trz',
    },
    MotionStep.Kind.Rotation : {
        Axis.X : 'rotx',
        Axis.Y : 'roty',
        Axis.Z : 'rotz',
    },
}


def _isIdentity(expr):
    return expr.expr == expr.arg.symbol

def expressionToMotionDSLSnippet(expr):
    if isinstance(expr, float):
        return expr.__str__()
    if expr.constant() :
        if expr.arg.name == "pi":
            return expr.__str__()
        if _isIdentity(expr):
            return "c:{}:{}".format(expr.arg.name, expr.evalf())
        else:
            cref = "c:" + expr.arg.name
            return expr.__str__().replace(expr.arg.name, cref)
    elif isinstance(expr.arg, Parameter):
        pref = "p:" + expr.arg.name
        if expr.arg.defaultValue is not None:
            pref = pref + f"[{expr.arg.defaultValue}]"
        return expr.__str__().replace(expr.arg.name, pref)
    else:
        return expr.__str__()

def poseSpecToMotionDSLSnippet(poseSpec):
    tgtF = poseSpec.pose.target
    refF = poseSpec.pose.reference
    steps = []
    for seq in poseSpec.motion.sequences :
        for step in seq.steps :
            steps.append("{step}({amount})".format(
                step   = __ser_map[step.kind][step.axis],
                amount = expressionToMotionDSLSnippet(step.amount))
            )
    ret = '{ref} -> {tgt} : {path}'.format(ref=refF.name, tgt=tgtF.name, path=" ".join(steps))
    return ret


def snippetToPoseSpec(text):
    preable = '''Model __tmp__\nConvention = currentFrame\n\n'''
    model_text = preable + text
    poses = toPosesSpecification( dsl.modelFromText(model_text) )
    return poses.poses[0]



'''
Concrete representations of coordinate transforms.

A concrete representation is typically a matrix, like a 4x4 matrix for
homogeneous coordinate vectors.

Please refer to the docs of the module `kgprim.ct.repr.mxrepr`; the other
modules are for internal use.
'''



from collections import OrderedDict

from kgprim.motions import MotionStep
import kgprim.values as numeric_argument


class TransformMetadata:
    '''
    Metadata of a single coordinate transform model.

    A metadata object contains the set of variables, parameters and constants
    which are part of the transform, and also two flags to tell whether the
    transform is parametric or constant.
    '''

    def __init__(self, coordinateTransform, customName=None):
        self.vars, self.pars, self.consts =\
                               symbolicArgumentsOf(coordinateTransform)
        self.ct = coordinateTransform
        self.parametric = (len(self.pars)>0)
        self.constant   = (len(self.vars)==0 and len(self.pars)==0)
        self._name      = customName or str(self.ct)

    @property
    def name(self): return self._name

    @property
    def is_parametric(self): return self.parametric

    @property
    def is_constant(self): return self.constant

    @property
    def is_dependent(self):
        '''Tell whether this transform depends on any Variable.'''
        return len(self.vars)>0

    @property
    def variables(self):
        '''
        Return a view of the set of Variables this transform depends on.
        The set is ordered. See the description of the keys of the dictionaries
        in `symbolicArgumentsOf()`.
        '''
        return self.vars.keys()

    @property
    def parameters(self):
        '''
        Return a view of the set of Parameters this transform depends on.
        The set is ordered. See the description of the keys of the dictionaries
        in `symbolicArgumentsOf()`.
        '''
        return self.pars.keys()

    @property
    def constants(self):
        '''
        Return a view of the set of Constants that appear in this transform.
        The set is ordered. See the description of the keys of the dictionaries
        in `symbolicArgumentsOf()`.
        '''
        return self.consts.keys()

    @property
    def variable_expressions(self):
        '''
        The dictionary with the variables and the corresponding expressions
        appearing in this transform.
        See `symbolicArgumentsOf()` for the format of the dictionary.
        '''
        return self.vars

    @property
    def parameter_expressions(self):
        '''
        The dictionary with the parameters and the corresponding expressions
        appearing in this transform.
        See `symbolicArgumentsOf()` for the format of the dictionary.
        '''
        return self.pars

    @property
    def constant_expressions(self):
        '''
        The dictionary with the constants and the corresponding expressions
        appearing in this transform.
        See `symbolicArgumentsOf()` for the format of the dictionary.
        '''
        return self.consts


class TransformsModelMetadata:
    '''
    Metadata for a whole transforms-model (that is, a set of transforms)
    '''

    def __init__(self, ctmodel, tfNamesMap=None):
        '''
        Arguments
          - `ctmodel`: the coordinate-transforms container, an instance of
            `models.CTransformsModel`
          - `tfNamesMap`: an optional dictionary with custom names for each
            transform in the given model. The dictionary is keyed with tuples
            with the left and right-frame (in this order) of the transform to
            be named.
        '''

        variables    = OrderedDict()
        parameters   = OrderedDict()
        constants    = OrderedDict()
        transforms = []
        def add(container, argument, expressions):
            if argument not in container.keys():
                container[ argument ] = expressions
            else :
                container[ argument ].update( expressions )

        names = tfNamesMap or {}
        for transf in ctmodel.transforms :
            name  = names.get( (transf.leftFrame, transf.rightFrame) ) # 'get' defaults to None
            tinfo = TransformMetadata(transf, customName=name)
            for var, expressions in tinfo.vars.items() :
                add( variables, var, expressions )
                #rotOrTr(tinfo, var)
            for par, expressions in tinfo.pars.items() :
                add( parameters, par, expressions )
#                     if rotOrTr(tinfo, par) : #it is a rotation
#                         rotationPars.append( par )
#                     else :
#                         translationPars.append( par )
            for cc, expressions in tinfo.consts.items() :
                add( constants, cc, expressions )
                #    rotOrTr(tinfo, cc)
            transforms.append( tinfo )

        self.ctModel = ctmodel
        self.transformsMetadata = transforms
        self.variables  = variables
        self.parameters = parameters
        self.constants  = constants

    @property
    def name(self):   return self.ctModel.name

    def isParametric(self):
        return len(self.parameters)>0



class UniqueExpression:
    '''
    Wraps a `kgprim.values.Expression` after stripping any leading minus.
    Expressions like `2*x` and `-2*x` would lead to two instances
    of this class that compare equal.

    Arguments:
    - `arg` a `kgprim.ct.models.PrimitiveCTransform` whose motion-step has an
      expression as argument; alternatively, a `kgprim.motions.MotionStep`
      object right away
    '''

    def __init__(self, arg ):
        try:
            src_expr = arg.amount # works for both PrimitiveCTransform and MotionStep

            # Sympy specifics here... might not be very robust...
            # Essentially we want to isolate the same Expression but without any
            # '-' in front
            # We rely on the fact that the sympy epressions coming from an input
            # geometry model are not more complicated than 'coefficient * symbol'.
            # Explicitly passing 'symbol' makes it work also when the
            # coefficient is a float rather than rational (e.g. "0.5*pi" as
            # opposed to "pi/2").

            (mult, rest) = src_expr.expr.as_coeff_mul( src_expr.arg.symbol )
            if len(rest) > 1 :
                raise RuntimeError("Could not separate the coefficient in expression {}".format(src_expr.expr))
            # we assume there is only one term other than the multiplier
            sympynew = abs(mult) * rest[0]
            expression = numeric_argument.Expression(argument=src_expr.arg, sympyExpr=sympynew)

            self.expression = expression
            self.rotation   = (arg.kind == MotionStep.Kind.Rotation)
        except AttributeError as e:
            raise ValueError("Incompatible argument type given to UniqueExpression()") from e

    @property
    def symbolicExpr(self):
        '''The underlying Sympy expression'''
        return self.expression.expr

    def isRotation(self): return self.rotation

    def isIdentity(self):
        return (self.expression.arg.symbol == self.expression.expr)

    def __eq__(self, rhs):
        return (self.expression == rhs.expression)
    def __hash__(self) :
        return 7*hash(self.expression)


def symbolicArgumentsOf(coordinateTransform):
    '''
    The variables, parameters and constants the given transform depends on.

    The argument must be a `ct.models.CoordinateTransform` instance (or a
    `ct.models.PrimitiveCTransform`).

    The function returns three ordered dictionaries, with keys being
    respectively the variables, parameters and constants of the given transform
    (`kgprim.values.Variable`, `kgprim.values.Parameter`, and
    `kgprim.values.Constant`). Occurrences of pi and raw floating point values
    are never included.
    The keys are stored in the same order as they appear in the transform (e.g.
    if the transform is a rotation of r radians followed by a translation of t
    meters, r will appear before t).

    The values in the dictionaries are also sorted containers, with all the
    unique expressions having the corresponding key as an argument. Expressions
    differing only for the sign are considered the same.
    For example, if the transform is defined as `rotx(2r) roty(3r) rotz(-2r)`,
    the container corresponding to `r` will contain `2r` and `3r`.
    '''
    varss = OrderedDict()
    pars  = OrderedDict()
    consts= OrderedDict()
    for pct in coordinateTransform.primitives :
        if isinstance(pct.amount, numeric_argument.Expression) :
            arg = pct.amount.arg
            rtexpr = UniqueExpression(pct)

        # Use ordered dictionaries with no values to emulate sorted sets.
        # Iteration over a dictionary itself (no keys() nor values() ) defaults
        # to iteration over the keys, so it would appear to be an ordered
        # sequence, with no duplicates, as I want.
            if isinstance(arg, numeric_argument.Variable) :
                if arg not in varss : varss[ arg ] = OrderedDict()
                varss.get( arg ) [rtexpr] = None
            elif isinstance(arg, numeric_argument.Parameter) :
                if arg not in pars : pars[ arg ] = OrderedDict()
                pars.get( arg ) [rtexpr] = None
            elif isinstance(arg, numeric_argument.Constant) :
                if arg not in consts : consts[ arg ] = OrderedDict()
                consts.get( arg ) [rtexpr] = None
    return varss, pars, consts









from enum import Enum
import kgprim.ct.models as ctmodels
import kgprim.ct.repr.mxcommon as common


class CoordinatesConvention(Enum):
    rotationOnTop = 0
    translationOnTop = 1

class VectorType(Enum):
    motion = 0
    force  = 1

block_bottom_left = (slice(3,6),slice(0,3))
block_bottom_right= (slice(3,6),slice(3,6))
block_top_right   = (slice(0,3),slice(3,6))
block_top_left    = (slice(0,3),slice(0,3))


def __rotx__a_x_b(mx, s, c):
    common.__rotx__a_x_b(mx, s, c)
    #common.__rotx__a_x_b(mx[block_bottom_right], s, c) # this does not work with SymPy, the slice does not act as a view
    mx[block_bottom_right] = mx[block_top_left]  # this one instead works

def __rotx__b_x_a(mx, s, c):
    common.__rotx__b_x_a(mx, s, c)
    mx[block_bottom_right] = mx[block_top_left]

def __roty__a_x_b(mx, s, c):
    common.__roty__a_x_b(mx, s, c)
    mx[block_bottom_right] = mx[block_top_left]

def __roty__b_x_a(mx, s, c):
    common.__roty__b_x_a(mx, s, c)
    mx[block_bottom_right] = mx[block_top_left]

def __rotz__a_x_b(mx, s, c):
    common.__rotz__a_x_b(mx, s, c)
    mx[block_bottom_right] = mx[block_top_left]

def __rotz__b_x_a(mx, s, c):
    common.__rotz__b_x_a(mx, s, c)
    mx[block_bottom_right] = mx[block_top_left]



'''
The skew symmetric matrix associated with the vector (x,y,z)
         0   -z    y
         z    0   -x
        -y    x    0
'''
def __skew_x(mx, x):
    mx[1,2] = -x
    mx[2,1] =  x
def __skew_y(mx, y):
    mx[0,2] =  y
    mx[2,0] = -y
def __skew_z(mx, z):
    mx[0,1] = -z
    mx[1,0] =  z

def __set_skew(mx, length, blockSlice, whichSkew):
    # we need to do all these copies of subblocks because passing slices as r/w
    # views of the original matrix does NOT work in SymPy
    # With NumPy, the following would be sufficient:
    #   whichSkew( mx[ blockSlice ], length)

    block = mx[ blockSlice ]
    whichSkew(block, length)
    mx[ blockSlice ] = block



__tr_block = {
    CoordinatesConvention.rotationOnTop : {
        VectorType.motion : block_bottom_left,
        VectorType.force  : block_top_right
    },
    CoordinatesConvention.translationOnTop : {
        VectorType.motion : block_top_right,
        VectorType.force  : block_bottom_left
    }
}


onRight = ctmodels.TransformPolarity.movedFrameOnTheRight
onLeft  = ctmodels.TransformPolarity.movedFrameOnTheLeft

X = common.X
Y = common.Y
Z = common.Z


def _tr_matrix_setters(coordinates_convention):
    def matrix_setters(vector_kind):
        matrix_block     = __tr_block[coordinates_convention][vector_kind]
        return {
            X : {
                onRight : (lambda mx, length :  __set_skew(mx,  length, matrix_block, __skew_x)),
                onLeft  : (lambda mx, length :  __set_skew(mx, -length, matrix_block, __skew_x))
            },
            Y : {
                onRight : (lambda mx, length :  __set_skew(mx,  length, matrix_block, __skew_y)),
                onLeft  : (lambda mx, length :  __set_skew(mx, -length, matrix_block, __skew_y))
            },
            Z : {
                onRight : (lambda mx, length :  __set_skew(mx,  length, matrix_block, __skew_z)),
                onLeft  : (lambda mx, length :  __set_skew(mx, -length, matrix_block, __skew_z))
            }
        }
    return matrix_setters




# Pure rotation matrices for spatial vectors are the same regardless of the
# type of vector (motion or force), and regardless of the convention
setRot = {
    X : {
        onRight : __rotx__a_x_b,
        onLeft  : __rotx__b_x_a
    },
    Y : {
        onRight : __roty__a_x_b,
        onLeft  : __roty__b_x_a
    },
    Z : {
        onRight : __rotz__a_x_b,
        onLeft  : __rotz__b_x_a
    },
}


class SpatialCoordinatesMixin:
    def __init__(self, spatialCoordinatesConvention=CoordinatesConvention.rotationOnTop, **kwds):
        super().__init__(**kwds)
        self.coordinatesConvention = spatialCoordinatesConvention
        self.tr_matrix_setters = _tr_matrix_setters(spatialCoordinatesConvention)

    @property
    def matrixSize(self):
        return 6

    def setRotation(self, axis, polarity, mx, s, c):
        setRot[axis][polarity](mx, s, c)


class MotionVectorMixin(SpatialCoordinatesMixin):
    def __init__(self, **kwds):
        super().__init__(**kwds)
        self.translation_setters = self.tr_matrix_setters(VectorType.motion)

    def setTranslation(self, axis, polarity, mx, length):
        self.translation_setters[axis][polarity](mx, length)


class ForceVectorMixin(SpatialCoordinatesMixin):
    def __init__(self, **kwds):
        super().__init__(**kwds)
        self.translation_setters = self.tr_matrix_setters(VectorType.force)

    def setTranslation(self, axis, polarity, mx, length):
        self.translation_setters[axis][polarity](mx, length)




from enum import Enum

class CoordinateTransformPlaceholder:
    def __init__(self, leftFrame, rightFrame):
        self.leftF  = leftFrame
        self.rightF = rightFrame
    @property
    def leftFrame(self): return self.leftF
    @property
    def rightFrame(self): return self.rightF


class CoordinateTransform:
    '''
    A model of the transform from `rightFrame` coordinates to `leftFrame`
    coordinates.
    '''

    def __init__(self, leftFrame, rightFrame, primitives):
        self.rightF = rightFrame
        self.leftF  = leftFrame
        self._primitives = primitives

    @property
    def leftFrame(self): return self.leftF
    @property
    def rightFrame(self): return self.rightF
    @property
    def primitives(self):
        '''The list of `PrimitiveCTransform` this instance is comprised of.'''
        return self._primitives

    def __str__(self): return self.leftFrame.name + "_X_" + self.rightFrame.name
    def __repr__(self): return self.__str__()
    def __eq__(self, rhs):
        return (isinstance(rhs, CoordinateTransform)
               and self.leftF == rhs.leftF
               and self.rightF == rhs.rightF
               and self._primitives == rhs._primitives) # TODO this is too strict: I could have a different sequence which leads to the same transform
    def __hash__(self):
        return 3*hash(self.leftF) + 93*hash(self.rightF) + 31*hash(tuple(self.primitives))

class TransformPolarity(Enum):
    '''
    The attribute distinguishing the two coordinate transforms associated to the
    same rigid motion.

    Consider frame A and B, where B is taken to be moved with respect to A
    (e.g. by a rotation about z of alpha radians).
    The corresponding transform can take two forms: either it takes B
    coordinates and gives back A coordinates for the same vector, or the
    other way round. We call the first A_X_B and the second B_X_A.
    Assuming the matrix is always left multiplied with a column vector of
    coordinates (as in A_X_B * pB), this notation immediately tells which are
    the "input" and "output" frames.

    A_X_B has the `movedFrameOnTheRight` polarity, because B appears on the
    right hand side, B_X_A is instead `movedFrameOnTheLeft`.

    Technically, the polarity is an attribute of the relation between a rigid
    motion and a related coordinate transform.
    '''
    movedFrameOnTheRight = 0
    movedFrameOnTheLeft  = 1


class PrimitiveCTransform:
    '''
    A coordinate transform for a pure rotation/translation about/along one
    of the Cartesian axis.

    Instances of this class model the coordinate transforms associated with
    "primitive" motion steps, see `kgprim.motions.MotionStep`.
    '''

    strbits = {
        TransformPolarity.movedFrameOnTheRight : "R",
        TransformPolarity.movedFrameOnTheLeft  : "L"
    }

    def __init__(self, motion_step, polarity):
        self.motion   = motion_step
        self.polarity_= polarity

    @property
    def kind(self): return self.motion.kind
    @property
    def amount(self): return self.motion.amount
    @property
    def axis(self): return self.motion.axis
    @property
    def polarity(self): return self.polarity_

    @property
    def primitives(self):
        '''
        A list with the self element only.

        This method is implemented to emulate the behaviour of `CoordinateTransform`.
        '''
        return [self]

    def __str__(self):
        return 'ct_{pol}_{motion}'.format(
            pol=PrimitiveCTransform.strbits[self.polarity],
            motion=str(self.motion))
    def __repr__(self):
        return self.__str__()
    def __eq__(self, rhs):
        return (isinstance(rhs, PrimitiveCTransform)
               and self.motion == rhs.motion
               and self.polarity == rhs.polarity)
    def __hash__(self):
        return hash(self.motion) + 111*hash(self.polarity)


class CTransformsModel:
    '''
    A simple named container of coordinate transforms.
    '''
    def __init__(self, name, transforms):
        self.name = name
        self.transforms = transforms



import math
import numpy as np
import kgprim.values as myexpr

class NumericMixin:
    def sin(self, arg):
        return math.sin(arg)

    def cos(self, arg):
        return math.cos(arg)

    def identity(self):
        return np.identity( self.matrixSize )

    def matrix_repr(self, ct):
        mx = self.identity()
        for p in ct.primitives :
            amount = p.amount
            if isinstance(amount, myexpr.Expression) :
                try:
                    amount = amount.evalf()
                except RuntimeError as e:
                    raise RuntimeError('Could not compute the numeric matrix representation of the transform') from e
            mx = mx @ self.matrix[p.kind](p.axis, p.polarity, amount)
        return mx
    # self.matrix comes from the MatrixRepresentationMixin

